/*-------------------------------*-C++-*--------------------------------*
 * File:	FXvt.cpp
 *----------------------------------------------------------------------*
 * Copyright (c) 2002      Gilles Filippini <pini@free.fr>
 *        - FXvt
 *
 * This file is part of FXvt.
 *
 * FXvt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * FXvt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FXvt; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *----------------------------------------------------------------------*/

#include <fox/fx.h>
#include "../config.h"		/* NECESSARY */
#include "rxvtlib.h"
#include "rxvt.h"
#include "command.h"
#include "init.h"
#include "FXvt.h"
#include <vector>

using namespace fxvt;

FXDEFMAP(FXvt)FXvtMap[]={
  FXMAPFUNC(SEL_IO_READ, FXvt::ID_CMDFD, FXvt::onReadCmdFd),
  FXMAPFUNC(SEL_PAINT, FXvt::ID_CANVAS, FXvt::onPaintCanvas),
  FXMAPFUNC(SEL_CHORE, FXvt::ID_REFRESH_CHORE, FXvt::onRefreshChore),
  FXMAPFUNC(SEL_CONFIGURE, FXvt::ID_CANVAS, FXvt::onConfigureCanvas),
  FXMAPFUNC(SEL_FOCUSIN, 0, FXvt::onFocusIn),
  FXMAPFUNC(SEL_CHANGED, FXvt::ID_SCROLL, FXvt::onChangedScroll),

  FXMAPFUNC(SEL_KEYPRESS, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_KEYRELEASE, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_MOTION, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_ENTER, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_LEAVE, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_FOCUSIN, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_FOCUSOUT, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_SELECTION_LOST, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_SELECTION_GAINED, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_SELECTION_REQUEST, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_RAISED, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_CLIPBOARD_LOST, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_CLIPBOARD_GAINED, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
  FXMAPFUNC(SEL_CLIPBOARD_REQUEST, FXvt::ID_CANVAS, FXvt::onUnhandledXEvent),
};

FXIMPLEMENT(FXvt, FXHorizontalFrame, FXvtMap, ARRAYNUMBER(FXvtMap));

FXvt::FXvt(FXComposite * p, FXint nr, FXint nc, FXObject * tgt, FXSelector sel,
					 FXuint opts) :
	FXHorizontalFrame(p, opts, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
	nrows(nr), ncols(nc), refreshChore(NULL), configured(false), positionned(false)
{
	rxvt = (rxvt_t *)rxvt_calloc(1, sizeof(rxvt_t));
	rxvt_set_r(rxvt);		/* only assignment to _rxvt_vars */
	if (rxvt_init_vars(rxvt) < 0) {
		free(rxvt);
		rxvt = NULL;
		return;
	}
	
	rxvt->fxvt = this;
	rxvt->TermWin.ncol = ncols;
	rxvt->TermWin.nrow = nrows;

	target = tgt;
	message = sel;
	canvas = new FXCanvas(this, this, ID_CANVAS, LAYOUT_FILL_X|LAYOUT_FILL_Y);
	vscrollbar = new FXScrollbar(this, this, ID_SCROLL, SCROLLBAR_VERTICAL|LAYOUT_FILL_Y);
	/*
	 * Save and then give up any super-user privileges
	 * If we need privileges in any area then we must specifically request it.
	 * We should only need to be root in these cases:
	 *  1.  write utmp entries on some systems
	 *  2.  chown tty on some systems
	 */
	rxvt_privileges(rxvt, SAVE);
	rxvt_privileges(rxvt, IGNORE);
	
	rxvt_init_secondary(rxvt);
	
	rxvt_init_resources(rxvt);
	
#if (MENUBAR_MAX)
	rxvt_menubar_read(rxvt, rxvt->h->rs[Rs_menu]);
#endif
#ifdef HAVE_SCROLLBARS
	if (rxvt->Options & Opt_scrollBar)
		rxvt->scrollBar.state = 1;	/* set existence for size calculations */
#endif
}

void FXvt::create()
{
	FXHorizontalFrame::create();
	rxvt_Create_Windows(rxvt, getParent());

	rxvt_init_xlocale(rxvt);
	
	rxvt_scr_reset(rxvt);		/* initialize screen */
	rxvt_Gr_reset(rxvt);		/* reset graphics */
	
#ifdef DEBUG_X
	XSynchronize(rxvt->Xdisplay, True);
	XSetErrorHandler((XErrorHandler) abort);
#else
	XSetErrorHandler((XErrorHandler) rxvt_xerror_handler);
#endif
	
#ifdef HAVE_SCROLLBARS
	if (rxvt->Options & Opt_scrollBar)
		rxvt_Resize_scrollBar(rxvt);	/* create and map scrollbar */
#endif
#if (MENUBAR_MAX)
	if (menubar_visible(rxvt))
		XMapWindow(rxvt->Xdisplay, rxvt->menuBar.win);
#endif

	rxvt_init_env(rxvt);
	rxvt_init_command(rxvt);
	

	rxvt->TermWin.mapped = true;

	rxvt->scrollBar.win = vscrollbar->id();
	setFocus();
}

FXint FXvt::getDefaultWidth()
{
	return ncols * rxvt->TermWin.fwidth + 2 * rxvt->TermWin.int_bwidth + vscrollbar->getDefaultWidth();
}

FXint FXvt::getDefaultHeight()
{
	return nrows * rxvt->TermWin.fheight + 2 * rxvt->TermWin.int_bwidth;
}

#define EXECL_ARGS \
	std::vector<char *> args; \
	va_list ap; \
	va_start(ap, arg); \
	while(arg) { \
		args.push_back(const_cast<char *>(arg)); \
		arg = va_arg(ap, const char *); \
	} \
  args.push_back(NULL); \
  char * const * argv = args.begin();

int FXvt::exec(int type, const char * path, char * const argv[], char * const env[])
{
	rxvt->cmd_fd = rxvt_run_command(rxvt, type, path, argv, env);
	FXApp::instance()->addInput(rxvt->cmd_fd, INPUT_READ, this, ID_CMDFD);
	return rxvt->cmd_fd;
}

int FXvt::execShell()
{
	return exec();
}

int FXvt::execl(const char * path, const char * arg, ...)
{
	EXECL_ARGS;
	return exec(EXECV, path, argv);
}

int FXvt::execv(const char * path, char * const argv[])
{
	return exec(EXECV, path, argv);
}

int FXvt::execlp(const char * path, const char * arg, ...)
{
	EXECL_ARGS;
	return exec(EXECVP, path, argv);
}

int FXvt::execvp(const char * path, char * const argv[])
{
	return exec(EXECVP, path, argv);
}

int FXvt::execle(const char * path, char * const env[], const char * arg, ...)
{
	EXECL_ARGS;
	return exec(EXECVE, path, argv, env);
}

int FXvt::execve(const char * path, char * const argv[], char * const env[])
{
	return exec(EXECVE, path, argv, env);
}

bool FXvt::fillBuffer(void)
{
	int ret = read(rxvt->cmd_fd, buffer, 1024);
	if (ret > 0) {
		buffer_cur = buffer;
		buffer_end = buffer + ret;
		/*for (int i = 0; i < ret; i++)
			fprintf(stderr, "%04o ", buffer[i]);
			fprintf(stderr, "\n");*/
		return true;
	}
	else if (!ret || (errno == EIO)) {
		FXApp::instance()->removeInput(rxvt->cmd_fd, INPUT_READ);
		close(rxvt->cmd_fd);
	}
	return false;
}

char FXvt::getNextChar()
{
	++buffer_cur;
	while (buffer_cur == buffer_end)
		fillBuffer();
	return *buffer_cur;
}

void FXvt::processBuffer()
{
	// Taken from rxvt command.c source
	char ch, *str;
	int nlines, refreshnow;
	struct rxvt_hidden *h = rxvt->h;
	
	refreshnow = 0;
	for (; buffer_cur != buffer_end; buffer_cur++) {
		ch = *buffer_cur;
		
		if (ch >= ' ' || ch == '\t' || ch == '\n' || ch == '\r') {
			/* Read a text string from the input buffer */
			/*
			 * point `str' to the start of the string,
			 * decrement first since it was post incremented in rxvt_cmd_getc()
			 */
			for (str = buffer_cur, nlines = 0; buffer_cur != buffer_end; buffer_cur++) {
				ch = *buffer_cur;
				if (ch == '\n') {
					nlines++;
					h->refresh_count++;
					if (!(rxvt->Options & Opt_jumpScroll)
							|| (h->refresh_count >= (h->refresh_limit
																			 * (rxvt->TermWin.nrow - 1)))) {
						refreshnow = 1;
						break;
					}
				} else if (ch < ' ' && ch != '\t' && ch != '\r') {
					/* unprintable */
					buffer_cur--;
					break;
				}
	    }
	    rxvt_scr_add_lines(rxvt, str, nlines, (buffer_cur - str + ((buffer_cur == buffer_end) ? 0 : 1)));
			/*
			 * If there have been a lot of new lines, then update the screen
			 * What the heck I'll cheat and only refresh less than every page-full.
			 * the number of pages between refreshes is h->refresh_limit, which
			 * is incremented here because we must be doing flat-out scrolling.
			 *
			 * refreshing should be correct for small scrolls, because of the
			 * time-out
			 */
	    if (refreshnow) {
				refreshnow = 0;
				if ((rxvt->Options & Opt_jumpScroll)
						&& h->refresh_limit < REFRESH_PERIOD)
					h->refresh_limit++;
				rxvt_scr_refresh(rxvt, h->refresh_type);
	    }
			// We may have reached the buffer's end
			if (buffer_cur == buffer_end)
				break;
		}
		else {
	    switch (ch) {
	    case C0_ENQ:	/* terminal Status */
				if (h->rs[Rs_answerbackstring])
					rxvt_tt_write(rxvt,
												(const unsigned char *)h->rs[Rs_answerbackstring],
												(unsigned int)STRLEN(h->rs[Rs_answerbackstring]));
				else
					rxvt_tt_write(rxvt, (unsigned char *)VT100_ANS,
												(unsigned int)STRLEN(VT100_ANS));
				break;
	    case C0_BEL:	/* bell */
				rxvt_scr_bell(rxvt);
				break;
	    case C0_BS:		/* backspace */
				rxvt_scr_backspace(rxvt);
				break;
	    case C0_VT:		/* vertical tab, form feed */
	    case C0_FF:
				rxvt_scr_index(rxvt, UP);
				break;
	    case C0_SO:		/* shift out - acs */
				rxvt_scr_charset_choose(rxvt, 1);
				break;
	    case C0_SI:		/* shift in - acs */
				rxvt_scr_charset_choose(rxvt, 0);
				break;
	    case C0_ESC:	/* escape char */
				rxvt_process_escape_seq(rxvt);
				break;
				/* case 0x9b: */	/* CSI */
				/*  rxvt_process_csi_seq(r); */
	    }
		}
	}
	/* NOTREACHED */
}

void FXvt::updateScrollBar()
{
	int line = 1;
	int page = rxvt->TermWin.nrow - line;
	int range = rxvt->TermWin.nscrolled + page;
	int position = rxvt->TermWin.nscrolled - rxvt->TermWin.view_start;

	if (vscrollbar->getRange() != range)
		vscrollbar->setRange(range);
	if (vscrollbar->getPage() != page)
		vscrollbar->setPage(page);
	if (vscrollbar->getLine() != line)
		vscrollbar->setLine(line);
	if (vscrollbar->getPosition() != position)
		vscrollbar->setPosition(position);
}

void FXvt::scheduleRefresh()
{
	if (!refreshChore)
		refreshChore = FXApp::instance()->addChore(this, ID_REFRESH_CHORE);
}

long FXvt::onReadCmdFd(FXObject *, FXSelector, void *)
{
	if (fillBuffer()) {
		processBuffer();
		scheduleRefresh();
	}

	return 1;
}

long FXvt::onPaintCanvas(FXObject *, FXSelector, void * ptr)
{
  FXEvent * event = reinterpret_cast<FXEvent *>(ptr);
  FXRectangle & rect = event->rect;

#define BORDER rxvt->TermWin.int_bwidth
#define WIDTH canvas->getWidth()
#define HEIGHT canvas->getHeight()
#define VTWIDTH (rxvt->TermWin.width + 2 * BORDER)
#define VTHEIGHT (rxvt->TermWin.height + 2 * BORDER)
	FXDCWindow dc(canvas, event);
	dc.setForeground(rxvt->PixColors[Color_bg]);
	// dc.setForeground(fxcolorfromname("violet"));
	dc.setLineWidth(BORDER);
	dc.drawRectangle(BORDER / 2, BORDER / 2, VTWIDTH - BORDER, VTHEIGHT - BORDER);
	// Remaining wrap margin
	if ((WIDTH - VTWIDTH) || (HEIGHT - VTHEIGHT)) {
		// dc.setForeground(fxcolorfromname("orange"));
		dc.setLineWidth(1);
	}
	if (WIDTH - VTWIDTH)
		dc.fillRectangle(VTWIDTH, 0, WIDTH - VTWIDTH, HEIGHT); 
	if (HEIGHT - VTHEIGHT)
		dc.fillRectangle(0, VTHEIGHT, VTWIDTH, HEIGHT - VTHEIGHT);
#undef BORDER
#undef WIDTH
#undef HEIGHT
#undef VTWIDTH
#undef VTHEIGHT

	rxvt_scr_expose(rxvt, rect.x, rect.y, rect.w, rect.h, true);
	rxvt->h->want_refresh = true;

	return 1;
}

long FXvt::onConfigureCanvas(FXObject *, FXSelector, void * ptr)
{
	rxvt_resize_all_windows(rxvt, canvas->getWidth(), canvas->getHeight());
	// Send refresh screen (ESC v) to the hosted application.
	// Side effect: shouldn't be sent the very first time, otherwise prints
	// an ugly "^[v".
	if (configured)
		rxvt_tt_printf(rxvt, "%c%c", 0033, 0166);
	else
		configured = true;
	return 1;
}

long FXvt::onRefreshChore(FXObject *, FXSelector, void * ptr)
{
	if (rxvt->h->want_refresh)
		// <FIXME>
		// rxvt_scr_refresh(rxvt, rxvt->h->refresh_type);
		rxvt_scr_expose(rxvt, canvas->getX(), canvas->getY(), canvas->getWidth(), canvas->getHeight(), true);
  	// </FIXME>
	refreshChore = NULL;
	return 1;
}

long FXvt::onFocusIn(FXObject *, FXSelector, void * ptr)
{
	rxvt_process_x_event(rxvt, (dynamic_cast<FXvtApplication *>(FXApp::instance()))->getCurEvent());
	canvas->setFocus();
	return 1;
}

long FXvt::onChangedScroll(FXObject *, FXSelector, void * ptr)
{
	if (vscrollbar->getPosition() != rxvt->TermWin.nscrolled - rxvt->TermWin.view_start) {
		rxvt_scr_move_to(rxvt, vscrollbar->getPosition(), vscrollbar->getRange());
		scheduleRefresh();
	}
	return 1;
}

long FXvt::onUnhandledXEvent(FXObject *, FXSelector, void * ptr)
{
	rxvt_process_x_event(rxvt, (dynamic_cast<FXvtApplication *>(FXApp::instance()))->getCurEvent());
	scheduleRefresh();
	return 1;
}
