/*-------------------------------*-C++-*--------------------------------*
 * File:	FXvt.h
 *----------------------------------------------------------------------*
 * Copyright (c) 2002      Gilles Filippini <pini@free.fr>
 *        - FXvt
 *
 * This file is part of FXvt.
 *
 * FXvt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * FXvt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FXvt; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *----------------------------------------------------------------------*/

typedef rxvt_t;

namespace fxvt {
	class FXvtApplication
	{
		friend class FXvt;
	protected:
		virtual FXRawEvent * getCurEvent() = 0;
	};
	
	class FXvt : public FXHorizontalFrame
	{
		FXDECLARE(FXvt);
	protected:
		FXint nrows;
		FXint ncols;
		FXScrollbar * vscrollbar;
		FXCanvas * canvas;
		rxvt_t * rxvt;
		char buffer[1024];
		char * buffer_cur;
		char * buffer_end;
		FXChore * refreshChore;
		bool configured;
		bool positionned;
	protected:
		FXvt() {}
		int exec(int type = 0, const char * path = NULL,
						 char * const argv[] = NULL, char * const env[] = NULL);
		void scheduleRefresh();
	public:
		enum {
			ID_CMDFD = FXHorizontalFrame::ID_LAST,
			ID_CANVAS,
			ID_SCROLL,
			ID_REFRESH_CHORE,
			ID_LAST
		};
	protected:
		// Terminal output stream handling
		bool fillBuffer();
		void processBuffer();
		public: char getNextChar();
	public:
		// Build a new new vt widget, with default size set to nr*nc (characters)
		FXvt(FXComposite * p, FXint nr = 25, FXint nc = 80, FXObject * tgt = NULL, FXSelector sel = 0, FXuint opts = 0);

		// Standard FOX virtual methods
		virtual void create();
		virtual FXint getDefaultWidth();
		virtual FXint getDefaultHeight();

		// Exec API
		int execShell();
		int execv(const char * path, char * const argv[]);
		int execl(const char * path, const char * arg, ...);
		int execve(const char * path, char * const argv[], char * const env[]);
		int execle(const char * path, char * const env[], const char * arg, ...);
		int execvp(const char * path, char * const argv[]);
		int execlp(const char * path, const char * arg, ...);

		// Events handling
		long onFocusIn(FXObject *, FXSelector, void *);
		long onReadCmdFd(FXObject *, FXSelector, void *);
		long onPaintCanvas(FXObject *, FXSelector, void *);
		long onConfigureCanvas(FXObject *, FXSelector, void *);
		long onRefreshChore(FXObject *, FXSelector, void *);
		long onChangedScroll(FXObject *, FXSelector, void *);

		// Dirty hacks. Keep out!
		FXCanvas * getCanvas() { return canvas; }
		void updateScrollBar();		
		long onUnhandledXEvent(FXObject *, FXSelector, void *);
	};
};
